package edu.mit.csail.pag.recrash;

import java.util.LinkedHashMap;

import checkers.quals.Default;
import checkers.quals.Nullable;

/**
 * Class name and loading related utility
 * 
 * @author Sung
 * 
 */
/* @Default("checkers.quals.NonNull") */
public class GetClass {

	static LinkedHashMap<String, Class<?>> typeToClass = new LinkedHashMap<String, Class<?>>();
	/**
	 * The order of primitiveType and ShortType name should be the same
	 */

	private static final Class<?> primitiveTypes[] = { void.class, // VOID_TYPE;
			boolean.class, // BOOLEAN_TYPE;
			char.class, // CHAR_TYPE;
			byte.class, // BYTE_TYPE;
			short.class, // SHORT_TYPE;
			int.class, // INT_TYPE;
			float.class, // FLOAT_TYPE;
			long.class, // LONG_TYPE;
			double.class // DOUBLE_TYPE;
	};

	private static String primitiveShortTypeName[] = { "V", // VOID_TYPE;
			"Z", // BOOLEAN_TYPE;
			"C", // CHAR_TYPE;
			"B", // BYTE_TYPE;
			"S", // SHORT_TYPE;
			"I", // INT_TYPE;
			"F", // FLOAT_TYPE;
			"J", // LONG_TYPE;
			"D" // DOUBLE_TYPE;
	};

	private static String objectForprimitiveType[] = { "Void", // VOID_TYPE;
			"Boolean", // BOOLEAN_TYPE;
			"Character", // CHAR_TYPE;
			"Byte", // BYTE_TYPE;
			"Short", // SHORT_TYPE;
			"Integer", // INT_TYPE;
			"Float", // FLOAT_TYPE;
			"Long", // LONG_TYPE;
			"Double" // DOUBLE_TYPE;
	};

	private static String primitiveJavaLangTypes[] = { "java.lang.Void", // VOID_TYPE;
			"java.lang.Boolean", // BOOLEAN_TYPE;
			"java.lang.Character", // CHAR_TYPE;
			"java.lang.Byte", // BYTE_TYPE;
			"java.lang.Short", // SHORT_TYPE;
			"java.lang.Integer", // INT_TYPE;
			"java.lang.Float", // FLOAT_TYPE;
			"java.lang.Long", // LONG_TYPE;
			"java.lang.Double" // DOUBLE_TYPE;
	};

	/**
	 * Get corresponding object name (for primitive types). For generic object,
	 * we just return object name
	 * 
	 * @param typeName
	 * @return
	 */
	public static String getCorrespondingObjectName(String typeName) {
		// primitive type
		for (int i = 0; i < primitiveTypes.length; i++) {
			if (typeName.equals(primitiveTypes[i].getName())) {
				return objectForprimitiveType[i];
			}
		}

		return typeName;
	}

	/**
	 * Return a corresponding class from type name
	 * 
	 * @param typeName
	 *            (slash separated)
	 * @return
	 * @throws ClassNotFoundException
	 */
	public static/* @Nullable */
	Class<?> fromTypeName(String typeName) throws ClassNotFoundException {
		/* @Nullable */
		Class<?> result = typeToClass.get(typeName);
		if (result != null)
			return result;

		// primitive type
		for (Class<?> primitive : primitiveTypes) {
			if (typeName.equals(primitive.getName())) {
				typeToClass.put(typeName, primitive);
				return primitive;
			}
		}

		/* @Nullable */
		String nullableString = typeName.replaceAll("]", "");
		if (nullableString == null) {
			return null;
		}

		String arrayTransedTypeName = nullableString;
		String arrayPart = "";
		String typeNamePart = arrayTransedTypeName;

		int arrayPoint = arrayTransedTypeName.indexOf('[');
		if (arrayPoint != -1) {
			nullableString = arrayTransedTypeName.substring(arrayPoint);
			if (nullableString == null) {
				return null;
			}

			arrayPart = nullableString;

			nullableString = arrayTransedTypeName.substring(0, arrayPoint);
			if (nullableString == null) {
				return null;
			}
			typeNamePart = nullableString;
		}

		for (int i = 0; i < primitiveTypes.length; i++) {
			Class<?> primitive = primitiveTypes[i];
			if (typeNamePart.equals(primitive.getName())) {
				result = Class.forName(arrayPart + primitiveShortTypeName[i]);
				if (result != null) {
					typeToClass.put(typeName, result);
				}
				return result;
			}
		}

		// If it is object array, change it to
		// [Ljava/lang/Object;
		if (arrayPoint != -1) {
			typeNamePart = "L" + typeNamePart + ";";
		}

		// general object
		result = Class.forName(arrayPart + typeNamePart);
		if (result != null) {
			typeToClass.put(typeName, result);
		}
		return result;
	}

	/**
	 * Return a corresponding class from type name
	 * 
	 * @param typeName
	 * @return
	 * @throws ClassNotFoundException
	 */
	public static boolean isPrimitive(String typeName)
			throws ClassNotFoundException {
		// primitive type
		for (Class<?> primitive : primitiveTypes) {
			if (typeName.equals(primitive.getName())) {
				typeToClass.put(typeName, primitive);
				return true;
			}
		}

		return false;
	}

	/**
	 * Return a corresponding class from type name
	 * 
	 * @param typeName
	 * @return
	 * @throws ClassNotFoundException
	 */
	public static boolean isPrimitiveJavaLangType(String typeName)
			throws ClassNotFoundException {
		// primitive type
		for (String name : primitiveJavaLangTypes) {
			if (name.equals(typeName)) {
				return true;
			}
		}

		return false;
	}

	/**
	 * Check if the type matches with "D", "Z", ... (short type name)
	 * 
	 * @param typeName
	 * @return
	 * @throws ClassNotFoundException
	 */
	public static boolean isPrimitiveShortType(String typeName)
			throws ClassNotFoundException {
		// primitive type
		for (String name : primitiveShortTypeName) {
			if (name.equals(typeName)) {
				return true;
			}
		}

		return false;
	}

	/*
	 * "V", // VOID_TYPE; "Z", // BOOLEAN_TYPE; "C", // CHAR_TYPE; "B", //
	 * BYTE_TYPE; "S", // SHORT_TYPE; "I", // INT_TYPE; "F", // FLOAT_TYPE; "J", //
	 * LONG_TYPE; "D" // DOUBLE_TYPE;
	 */
	public static String descToSig(String desc) throws Exception {
		String sig = "";
		String currentArray = "";
		for (int i = 0; i < desc.length(); i++) {
			if (desc.charAt(i) == ')') {
				return sig;
			}

			if (sig.length() > 0) {
				sig += ",";
			}
			switch (desc.charAt(i)) {
			case '(':
			case 'V':
				continue;
			case '[':
				currentArray += "[]";
				continue;
			case 'C':
				sig += "char";
				break;
			case 'B':
				sig += "byte";
				break;
			case 'S':
				sig += "short";
				break;
			case 'I':
				sig += "int";
				break;
			case 'F':
				sig += "float";
				break;
			case 'J':
				sig += "long";

				break;
			case 'D':
				sig += "double";
				break;
			case 'Z':
				sig += "boolean";
				break;

			case 'L':
				int indexP = desc.indexOf(';', i);
				if (indexP == -1) {
					throw new Exception("Malformed signature description");
				}
				sig = desc.substring(i + 1, indexP);
				i = indexP;
				break;
			default:
				throw new Exception("Malformed signature description: "
						+ desc.charAt(i) + " from: " + desc);

			}

			sig += currentArray;
			currentArray = "";
		}

		return sig;
	}
}
