package edu.mit.csail.pag.recrash.generator;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Date;
import java.util.Set;

import edu.mit.csail.pag.recrash.GetClass;
import edu.mit.csail.pag.recrash.Logger;
import edu.mit.csail.pag.recrash.Util;
import edu.mit.csail.pag.recrash.tracer.TraceReader;
import edu.mit.csail.pag.recrash.tracer.item.MethodTraceItem;
import edu.mit.csail.pag.recrash.tracer.item.ClonedArgumentObject;
import edu.mit.csail.pag.recrash.tracer.item.StackTraceItem;

/**
 * Generating code trace stored by reCrash
 *
 * TODO: need to write the output to the file
 *
 * @author hunkim
 *
 */

public class CrashTestCaseGenerator {

	// initial support string
	public final String imports[] = { "import junit.framework.TestCase;",
			"import edu.mit.csail.pag.recrash.tracer.*;",
			"import edu.mit.csail.pag.recrash.tracer.item.*;",
			"import edu.mit.csail.pag.recrash.generator.*;" };

	/**
	 * Subclass (generated code and utility)
	 *
	 * @author hunkim
	 *
	 */
	static class GeneratedCode {
		String code;

		String className;

		void showCode() {
			System.out.println(code);
		}

		void toFile(String sourceDir, String packageName) throws IOException {
			assert (code != null && className != null && sourceDir != null && packageName != null);
			File javaFile = new File(sourceDir + File.separator
					+ packageName.replace('.', '/'), className + ".java");
			javaFile.getParentFile().mkdirs();
			FileWriter out = new FileWriter(javaFile);
			System.err.println("Generated test cases go to: " + javaFile);
			out.write(code + "\n");
			out.close();
		}

	}

	/**
	 * Generate code from traceItem
	 *
	 * @param packageName
	 * @return
	 * @throws ClassNotFoundException
	 */
	public GeneratedCode codeGeneration(String packageName, String traceFileName)
			throws IOException, ClassNotFoundException {
		if (traceFileName == null) {
			traceFileName = Util.TRACE_FILE;
		}

		StackTraceItem stackTraceItem = TraceReader.readTrace(traceFileName);
		if (stackTraceItem == null || stackTraceItem.size() == 0) {
			Logger.info("No methods to generate tests. Bye!!");
			return null;
		}
		// If object is stored as string, lets load them
		stackTraceItem.unMarshallMethodItems();

		GeneratedCode generatedCode = new GeneratedCode();

		// Package and starting stuff
		String generatedCodeString = "package " + packageName + ";\n\n";

		// print out imports
		for (int i = 0; i < imports.length; i++) {
			generatedCodeString += imports[i] + "\n";
		}

		String programNameVersionNumber = stackTraceItem
				.getProgramNameVersionNumber();
		if (programNameVersionNumber == null) {
			programNameVersionNumber = "noName";
		}

		generatedCodeString += "\n";

		generatedCodeString += "// generated by reCrash on " + new Date()
				+ "\n";
		generatedCodeString += "// Created on "
				+ TraceReader.getTraceFileDate() + " for "
				+ programNameVersionNumber + " JVM: "
				+ System.getProperty("java.vm.version") + "\n";

		File traceFile = new File(traceFileName);

		generatedCode.className = "Recrash_"
				+ traceFile.getName().replaceAll("\\W+", "_");

		// get Stack trace
		generatedCodeString += generateStackTrace(generatedCode, stackTraceItem);

		generatedCodeString += "\n";
		generatedCodeString += "public class " + generatedCode.className
				+ " extends TestCase {\n";

		String absoluteTraceFileName = traceFile.getAbsolutePath();
                // Presumes that backslashes don't appear in filenames
                // except as Windows path separators.
		String quotedTraceFileName = absoluteTraceFileName.replace("\\", "\\\\");

		// We need to setup, since a test can change the status of loaded object
		generatedCodeString += "\tpublic void setUp() throws Exception {\n"
				+ "\t\tTraceReader.readTrace(\"" + quotedTraceFileName
				+ "\");\n\t}\n\n";

		for (int s = stackTraceItem.size() - 1; s >= 0; s--) {
			MethodTraceItem traceItem = stackTraceItem.get(s);

			// failed to de-serialize?
			if (traceItem == null) {
				continue;
			}

			// Set trace item to get real value
			TraceReader.setMethodTraceItem(s);

			generatedCodeString += "\n\tpublic void test_"
					+ traceItem.getMethodName().replaceAll("\\W+", "_") + "_"
					+ s + "() throws Throwable {\n";
			generatedCodeString += "\t\tTraceReader.setMethodTraceItem(" + s
					+ ");\n";

			// It is a receiver, this
			ClonedArgumentObject thisObject = traceItem.getArgumentObject(0);

			boolean useProxy = false;

			// if it is not static
			if (thisObject != null) {
				// We need to use proxy if a object is not public or inner class
				useProxy = !thisObject.isPublic();

				String readingThisString = thisObject.getTypeName()
						+ " thisObject  = (" + thisObject.getTypeName()
						+ ") TraceReader.readObject(0);\n";

				if (useProxy) {
					generatedCodeString += "\t\t//" + readingThisString;
					generatedCodeString += "\t\tObject thisObject = TraceReader.readObject(0);\n";
				} else {
					generatedCodeString += "\t\t" + readingThisString;
				}
			}

			// Get set of all static objects
			if (false) {
				Set<String> staticClassSet = traceItem.getStaticClassNameSet();
				for (String staticClassName : staticClassSet) {
					generatedCodeString += "\n\t\t// load static objects\n";
					generatedCodeString += "\t\t" + staticClassName + " "
							+ staticClassName.replaceAll("\\.", "_") + " = ("
							+ staticClassName + ") TraceReader.readStatic (\""
							+ staticClassName + "\");\n";
				}
			}

			// reading arguments from trace
			generatedCodeString += "\n\t\t// load arguments\n";
			for (int i = 1; i < traceItem.getNumberOfArguments(); i++) {
				generatedCodeString += generateCodeForLoadingArgument(
						traceItem, i);
			}

			if (useProxy) {
				generatedCodeString += "\n\t\t// Collect argument types for proxy call";
				generatedCodeString += "\n\t\tString classTypeNames[] = {";

				for (int i = 1; i < traceItem.getNumberOfArguments(); i++) {
					if (i != 1) {
						generatedCodeString += ",\n\t\t\t\t";
					}
					ClonedArgumentObject argumentObject = traceItem
							.getArgumentObject(i);

					generatedCodeString += "\"" + argumentObject.getTypeName()
							+ "\"";
				}
				generatedCodeString += "};\n";
			}

			// call method
			String arguments = getArgumentListString(traceItem);
			String callMethodString = null;

			if (thisObject != null) {
				callMethodString = "thisObject."
						+ traceItem.getShortMethodName() + "(" + arguments
						+ ");";
			} else {
				callMethodString = traceItem.getMethodName() + "(" + arguments
						+ ");";
			}

			if (useProxy) {
				generatedCodeString += "\n\t\t// " + callMethodString;
				generatedCodeString += "\n\t\t// Method invocation using proxy";
				generatedCodeString += "\n\t\tnew ProxyObject(thisObject).call(\""
						+ traceItem.getShortMethodName() + "\", classTypeNames";

				if (arguments.length() > 0) {
					generatedCodeString += " ,";
				}
				// passing arguments
				generatedCodeString += arguments;

				// closing method call
				generatedCodeString += ");\n";
			} else {
				generatedCodeString += "\n\t\t// Method invocation";
				generatedCodeString += "\n\t\t" + callMethodString + "\n";
			}

			// closing method and class
			generatedCodeString += "\t}\n";

		}
		generatedCodeString += "}\n";

		generatedCode.code = generatedCodeString;
		return generatedCode;
	}

	/**
	 * Get argument list
	 *
	 * ex: arg1, arg2, ....
	 *
	 * @param traceItem
	 * @return
	 */
	private String getArgumentListString(MethodTraceItem traceItem) {
		String generatedCodeString = "";
		for (int i = 1; i < traceItem.getNumberOfArguments(); i++) {
			if (i != 1) {
				generatedCodeString += ", ";
			}
			generatedCodeString += Util.makeArgumentName(i);
		}
		return generatedCodeString;
	}

	/**
	 * Code for a parameter loading
	 *
	 * @param traceItem
	 * @param argOrder
	 * @return
	 * @throws ClassNotFoundException
	 */
	private String generateCodeForLoadingArgument(MethodTraceItem traceItem,
			int argOrder) throws ClassNotFoundException {
		String generatedCodeString = "";

		ClonedArgumentObject argumentObject = traceItem
				.getArgumentObject(argOrder);

		assert (argumentObject != null);

		String argumentName = Util.makeArgumentName(argOrder);

		// see if we can use the argument
		// ### FIXME: what happen if the type is non-public
		boolean canWecallDirect = argumentObject.isPublic();
		canWecallDirect = true;

		String argumentTypeName = "Object";

		if (canWecallDirect) {
			argumentTypeName = argumentObject.getTypeName();
			argumentTypeName = Util.getShortTypeName(argumentTypeName);
		}

		String argumentValue = "" + argumentObject.getObject();
		if (argumentValue.length() > 50) {
			argumentValue = argumentValue.substring(0, 49);
		}

		argumentValue = argumentValue.replace('\n', ' ');
		argumentValue = argumentValue.replace('\r', ' ');

		assert (argumentObject != null);

		// It can be representable by values
		if (argumentObject.getObject() == null
				|| GetClass.isPrimitive(argumentTypeName)
				|| GetClass.isPrimitiveJavaLangType(argumentTypeName)) {
			generatedCodeString += "\t\t" + argumentTypeName + " "
					+ argumentName + " = " + argumentValue + ";\n";
		} else if (argumentTypeName.endsWith("java.lang.String")) {
			generatedCodeString += "\t\t" + argumentTypeName + " "
					+ argumentName + " = \"" + argumentValue + "\";\n";
		} else {
			generatedCodeString += "\t\t// " + argumentTypeName + " "
					+ argumentName + " = " + argumentValue + ";\n";
			generatedCodeString += "\t\t" + argumentTypeName + " "
					+ argumentName + " = ";
			if (canWecallDirect) {
				generatedCodeString += "("
						+ GetClass.getCorrespondingObjectName(argumentTypeName)
						+ ")";
			}
			generatedCodeString += "TraceReader.readObject(" + argOrder
					+ ");\n";
		}

		return generatedCodeString;
	}

	/**
	 * Prints out stack trace as comments
	 *
	 * @param generatedCode
	 * @param stackTraceElements
	 * @return
	 */
	private String generateStackTrace(GeneratedCode generatedCode,
			StackTraceItem stackTraceItem) {

		Throwable exception = stackTraceItem.getCrashedException();

		if (exception == null) {
			return "\n//-----No stack trace ------------\n";
		}

		StackTraceElement[] stackTraceElements = exception.getStackTrace();
		String generatedCodeString = "\n//-----Stack trace: "
				+ exception.toString() + " ------------\n";
		for (int i = 0; i < stackTraceElements.length; i++) {
			generatedCodeString += "//" + stackTraceElements[i] + "\n";
		}

		Throwable cause = exception.getCause();
		if (cause == null) {
			return generatedCodeString;
		}

		stackTraceElements = cause.getStackTrace();
		generatedCodeString += "\n//----- Caused by: " + cause.toString()
				+ " ------------\n";
		for (int i = 0; i < stackTraceElements.length; i++) {
			generatedCodeString += "//" + stackTraceElements[i] + "\n";
		}

		return generatedCodeString;
	}

	static public void main(String args[]) throws IOException,
			ClassNotFoundException {
		String packageName = "generated_tests";

		if (args.length < 2) {
			System.out
					.println("crashTestGenerator trace_file [out_dir] [package_name]");
			System.exit(0);
		}

		GeneratedCode generetedCode = new CrashTestCaseGenerator()
				.codeGeneration(packageName, args[0]);
		if (generetedCode == null) {
			return;
		}
		// generated code output directory
		// usually your workspace/project/src
		if (args.length == 2) {
			generetedCode.toFile(args[1], packageName);
		}

		generetedCode.showCode();
	}
}
