package edu.mit.csail.pag.recrash.tracer;

import java.util.Hashtable;

import checkers.quals.Default;

import edu.mit.csail.pag.recrash.Logger;
import edu.mit.csail.pag.recrash.tracer.item.StackTraceItem;

/**
 * Make a thread safe tracer. We keep a trace item for each tread in stack.
 * Using stack is important:
 * 
 * <pre>
 * foo() { 
 *   // trace
 *   bar()
 *   // * crash point
 * }
 * 
 * bar() {
 *   // trace
 * }
 * </pre>
 * 
 * If we keep only the latest trace, the following crash. The crash point is
 * foo, not bar.
 * 
 * 
 * @author hunkim
 * 
 */

/*@Default("checkers.quals.NonNull")*/
public class ThreadHandler {
	// Hold a trace item stack per thread
	static private Hashtable<Long, StackTraceItem> traceItermPerThread = new Hashtable<Long, StackTraceItem>();
	static private StackTraceItem oneStackTrackItem = new StackTraceItem();

	static boolean useRealThreadId = true;
	static String programNameAndVersion = null;

	/**
	 * Entering a new method: create new traceItem and push
	 * 
	 * @param methodName
	 * @param doCloneObject 
	 */
	static public int enterMethod(String methodName, Object objects[],
			String types[], boolean[] doCloneObject) {
		StackTraceItem stackItem = getCurrentItem();

		return stackItem.enterMethod(methodName, objects, types, doCloneObject);
	}

	/**
	 * Exiting a method: pop the method
	 * 
	 * @param methodName
	 */
	static public void exitMethod(String methodName, int nounceNumber) {
		StackTraceItem stackItem = getCurrentItem();
		if (stackItem == null) {
			return;
		}

		stackItem.exitMethod(methodName, nounceNumber);
	}

	/**
	 * Get stack of current thread. We need to store whole stack to generated
	 * layered test cases
	 * 
	 * @return a stack trace item for thread
	 */
	public static StackTraceItem getCurrentItem() {

		/*
		if (!useRealThreadId) {
			return oneStackTrackItem;
		}
		*/

		Long threadId = Thread.currentThread().getId();
		StackTraceItem item = traceItermPerThread.get(threadId);
		if (item == null) {
			if (Logger.verbose) {
				Logger.info("Create a new stack for thread: " + threadId);
			}
			item = new StackTraceItem();
			traceItermPerThread.put(threadId, item);
		}

		return item;
	}

	public static void setProgramNameVersionNumber(String nameAndNumber) {
		programNameAndVersion = nameAndNumber;
	}

	public static String getProgramNameAndVersion() {
		return programNameAndVersion;
	}
}
