package edu.mit.csail.pag.recrash.tracer.item;

import java.util.ArrayList;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;
import java.util.Stack;

import checkers.quals.*;

import edu.mit.csail.pag.recrash.Logger;
import edu.mit.csail.pag.recrash.tracer.Serializer;
import edu.mit.csail.pag.recrash.tracer.ThreadHandler;

/**
 * This extends stack and holds Method items per thread Provide some wrappers
 * 
 * @author hunkim
 * 
 */
/* @Default("checkers.quals.NonNull") */
public class StackTraceItem extends Stack<MethodTraceItem> {
	int nounceNumber = 0;

	// decide if we have to serialize method items one by one
	private boolean serializeOneByOne;

	/* @Nullable */List<byte[]> serializedMethodItemArrayList = null;
	/**
	 * 
	 */
	private static final long serialVersionUID = 1L;

	// current exception
	private Throwable crashedException;

	private String programNameAndVersion;

	/**
	 * Entering a new method: create new traceItem and push
	 * 
	 * @param methodName
	 * @param doCloneObject
	 */
	public int enterMethod(String methodName, Object objects[], String types[],
			boolean[] doCloneObject) {
		MethodTraceItem traceItem = new MethodTraceItem(methodName, objects,
				types, doCloneObject, nounceNumber);
		this.push(traceItem);
		return nounceNumber++;
	}

	/**
	 * Exiting a method: pop the method
	 * 
	 * @param methodName
	 */
	public void exitMethod(String methodName, int nounceNumber) {
		// synchronized (this) {
		// we need make sure that this stack is not empty before we pop-up
		while (!isEmpty()) {
			MethodTraceItem traceItem = pop();
			assert (traceItem != null);

			if (traceItem.equals(nounceNumber)) {
				break;
			}

			if (Logger.verbose) {
				Logger.error("Stack mismatch: looking for " + methodName
						+ " but found " + traceItem.getMethodName());
			}
		}
		// }
	}

	/**
	 * Current top traceItem in current thread
	 * 
	 * @return
	 */
	public MethodTraceItem getMethodTraceItem() {
		return this.peek();
	}

	/**
	 * Get stack with only the last method trace item in order to save trace
	 * size layered test cases
	 * 
	 * @return
	 */
	public StackTraceItem getReducedStackItem(int n) {
		// If there is no exception info, don't reduce
		if (crashedException == null) {
			return this;
		}

		if (serializeOneByOne) {
			serializedMethodItemArrayList = new ArrayList<byte[]>();
		}

		// get a new stack and
		StackTraceItem newStackItem = new StackTraceItem();

		// cloning the stack item for new stack item
		newStackItem.crashedException = crashedException;
		newStackItem.nounceNumber = nounceNumber;
		newStackItem.serializedMethodItemArrayList = serializedMethodItemArrayList;

		// This will be set from ThreadHandler directly to reduce overhead
		newStackItem.programNameAndVersion = ThreadHandler
				.getProgramNameAndVersion();

		// #FIXME: reCrash shadow stack and back stack trace in Exception do not
		// match!
		// So we search methods in the shadow stack

		// We look for cause
		Throwable cause = crashedException;
		if (cause != null && cause.getCause() != null) {
			cause = cause.getCause();
		}

		StackTraceElement element[] = cause.getStackTrace();
		Set<String> lookingforMethodName = new LinkedHashSet<String>();

		// for (int i = element.length - 1; i >= 0; i--) {
		for (int i = 0; i < element.length; i++) {
			String fullMethodName = element[i].getClassName() + "."
					+ element[i].getMethodName();
			// check if we seek the method already
			if (lookingforMethodName.contains(fullMethodName)) {
				continue;
			}

			lookingforMethodName.add(fullMethodName);

			Logger.info("Look for: " + fullMethodName);
			addMethodToNewItem(fullMethodName, newStackItem);

			// Check if the method item size is grater than 0
			if (newStackItem.size() >= n
					|| (serializedMethodItemArrayList != null && serializedMethodItemArrayList
							.size() >= n)) {
				break;
			}
		}
		// return this;
		return newStackItem;
	}

	/**
	 * Find methodTraceItem using the name and add it to new place
	 * 
	 * @param fullMethodName
	 * @param newStackItem
	 */
	private void addMethodToNewItem(String fullMethodName,
			StackTraceItem newStackItem) {

		// it's a sequencial search
		for (int i = 0; i < size(); i++) {
			MethodTraceItem toAdd = this.get(i);

			if (!fullMethodName.equals(toAdd.getMethodName())) {
				continue;
			}

			// test if it can be serialized
			// If we can serialize it, we skip it
			try {
				Logger.debug("Serializing a object: " + fullMethodName);

				if (serializedMethodItemArrayList != null) {
					byte[] serializedByteArray = Serializer
							.toXMLZippedByteArray(toAdd);

					serializedMethodItemArrayList.add(serializedByteArray);
					Logger.info("Serialized: " + fullMethodName);
				} else {
					newStackItem.add(0, toAdd);
					Logger.info("Added: " + fullMethodName);
				}
			} catch (Exception e) {
				System.err.println("Skip method: " + toAdd.getMethodName()
						+ ": " + e);
			}
		}
	}

	/**
	 * Unmarshall each method items
	 */
	public void unMarshallMethodItems() {
		if (serializedMethodItemArrayList == null) {
			Logger.info("No serialized methods!");
			return;
		}

		for (byte[] serizliedString : serializedMethodItemArrayList) {
			Logger.info("Unmarshalling...");
			try {
				MethodTraceItem toAdd = (MethodTraceItem) Serializer
						.loadObject(serizliedString);
				push(toAdd);
			} catch (Throwable e) {
				Logger.error("Skip a method: " + e);
				continue;
			}
		}

		serializedMethodItemArrayList = null;
		System.gc();
	}

	/**
	 * Get exception
	 * 
	 * @return
	 */
	public Throwable getCrashedException() {
		return crashedException;
	}

	/**
	 * Store crash exception
	 * 
	 * @param crashedException
	 */
	public void setCrashedException(Throwable crashedException) {
		this.crashedException = crashedException;
	}

	public void addStatic(String fullyQualifiedStaticClassName,
			String serializedXMLString) {
		MethodTraceItem item = getMethodTraceItem();
		if (item != null) {
			item.addStatic(fullyQualifiedStaticClassName, serializedXMLString);
		}
	}

	/**
	 * Get programNameAndVersion This will be set when we reduce the
	 * StackTraceItem
	 * 
	 * @return
	 */
	public String getProgramNameVersionNumber() {
		return programNameAndVersion;
	}
}
